@@ -13,10 +13,10 @@
 import logging
 import os
 import signal
-import socket
 import sys
 import threading
 import time
+import typing
 
 from oslo_config import cfg
 
@@ -24,42 +24,37 @@ import cotyledon
 from cotyledon import _utils
 from cotyledon import oslo_config_glue
 
-if len(sys.argv) >= 3:
-    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-    s.connect(("127.0.0.1", int(sys.argv[2])))
-    if os.name == "posix":
-        stream = os.fdopen(s.fileno(), 'w')
-    else:
-        stream = s.makefile()
-    logging.basicConfig(level=logging.DEBUG, stream=stream)
-else:
-    logging.basicConfig(level=logging.DEBUG)
+
+if typing.TYPE_CHECKING:
+    from cotyledon import types
+
+logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
 
 LOG = logging.getLogger("cotyledon.tests.examples")
 
 # We don't want functional tests to wait for this:
-cotyledon.ServiceManager._slowdown_respawn_if_needed = lambda *args: True
+cotyledon.ServiceManager._slowdown_respawn_if_needed = lambda *args: True  # type: ignore[method-assign,assignment]
 
 
 class FullService(cotyledon.Service):
     name = "heavy"
 
-    def __init__(self, worker_id):
-        super(FullService, self).__init__(worker_id)
+    def __init__(self, worker_id: "types.WorkerId") -> None:
+        super().__init__(worker_id)
         self._shutdown = threading.Event()
-        LOG.error("%s init" % self.name)
+        LOG.error("%s init", self.name)
 
-    def run(self):
-        LOG.error("%s run" % self.name)
+    def run(self) -> None:
+        LOG.error("%s run", self.name)
         self._shutdown.wait()
 
-    def terminate(self):
-        LOG.error("%s terminate" % self.name)
+    def terminate(self) -> None:
+        LOG.error("%s terminate", self.name)
         self._shutdown.set()
         sys.exit(42)
 
-    def reload(self):
-        LOG.error("%s reload" % self.name)
+    def reload(self) -> None:
+        LOG.error("%s reload", self.name)
 
 
 class LigthService(cotyledon.Service):
@@ -70,14 +65,19 @@ class BuggyService(cotyledon.Service):
     name = "buggy"
     graceful_shutdown_timeout = 1
 
-    def terminate(self):
+    def terminate(self) -> None:
         time.sleep(60)
         LOG.error("time.sleep done")
 
 
+class BoomError(Exception):
+    pass
+
+
 class BadlyCodedService(cotyledon.Service):
-    def run(self):
-        raise Exception("so badly coded service")
+    def run(self) -> None:
+        msg = "so badly coded service"
+        raise BoomError(msg)
 
 
 class OsloService(cotyledon.Service):
@@ -88,19 +88,19 @@ class WindowService(cotyledon.Service):
     name = "window"
 
 
-def on_terminate():
+def on_terminate() -> None:
     LOG.error("master terminate hook")
 
 
-def on_terminate2():
+def on_terminate2() -> None:
     LOG.error("master terminate2 hook")
 
 
-def on_reload():
+def on_reload() -> None:
     LOG.error("master reload hook")
 
 
-def example_app():
+def example_app() -> None:
     p = cotyledon.ServiceManager()
     p.add(FullService, 2)
     service_id = p.add(LigthService, 5)
@@ -110,16 +110,15 @@ def example_app():
     p.run()
 
 
-def buggy_app():
+def buggy_app() -> None:
     p = cotyledon.ServiceManager()
     p.add(BuggyService)
     p.run()
 
 
-def oslo_app():
+def oslo_app() -> None:
     conf = cfg.ConfigOpts()
-    conf([], project='openstack-app', validate_default_values=True,
-         version="0.1")
+    conf([], project="openstack-app", validate_default_values=True, version="0.1")
 
     p = cotyledon.ServiceManager()
     oslo_config_glue.setup(p, conf)
@@ -127,7 +126,7 @@ def oslo_app():
     p.run()
 
 
-def window_sanity_check():
+def window_sanity_check() -> None:
     p = cotyledon.ServiceManager()
     p.add(LigthService)
     t = _utils.spawn(p.run)
@@ -136,18 +135,22 @@ def window_sanity_check():
     t.join()
 
 
-def badly_coded_app():
+def badly_coded_app() -> None:
     p = cotyledon.ServiceManager()
     p.add(BadlyCodedService)
     p.run()
 
 
-def exit_on_special_child_app():
+def exit_on_special_child_app() -> None:
     p = cotyledon.ServiceManager()
     sid = p.add(LigthService, 1)
     p.add(FullService, 2)
 
-    def on_dead_worker(service_id, worker_id, exit_code):
+    def on_dead_worker(
+        service_id: "types.ServiceId",
+        worker_id: "types.WorkerId",
+        exit_code: int,
+    ) -> None:
         # Shutdown everybody if LigthService died
         if service_id == sid:
             p.shutdown()
@@ -156,9 +159,8 @@ def exit_on_special_child_app():
     p.run()
 
 
-def sigterm_during_init():
-
-    def kill():
+def sigterm_during_init() -> None:
+    def kill() -> None:
         os.kill(os.getpid(), signal.SIGTERM)
 
     # Kill in 0.01 sec
@@ -168,5 +170,5 @@ def sigterm_during_init():
     p.run()
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     globals()[sys.argv[1]]()
